//
// Image-Klasse zur Vorlesung 
// Distributed Algorithms for Image and Video Processing
// (c) Stephan Kopf, Universitt Mannheim, 2009
//
#include<Image.h>

// Construction
Image::Image ()
{
	width  = 0;
	height = 0;
	band   = 3;
	data   = NULL;
}

// Construction
Image::Image (int w, int h, int b)
{
	width  = w;
	height = h;
	band   = b;
	data = NULL;
	resize (width, height, band);
}

// Construction
Image::Image (string &filename)
{
	width  = 0;
	height = 0;
	band   = 3;
	data   = NULL;

	read (filename);
}

// Destruction
Image::~Image ()
{
	freeData();
}


// Interface
void 
Image::resize  (int w, int h, int b)
{
	if (w!=width || h!=height || b!=band || data==NULL) {
		freeData();

		width  = w;
		height = h;
		band   = b;
		data = NULL;
		allocData();
	} // if
	return;
}

// Interface
void 
Image::clear (unsigned char value)
{
	for (int b=0; b<band; b++) {
		for (int h=0; h<height; h++) {
			for (int w=0; w<width; w++) {
				data[b][h][w] = value;
			}
		}
	}
	return;
}



// Interface
void 
Image::read (string &filename)
{
	// open file
    ifstream *is = NULL;
    is = new ifstream (filename.c_str(), std::ios::in | std::ios::binary);
    if (!is) {
		cerr << "error: cannot open file " << filename << endl;
		exit(1);
    } // if
    if (!is->is_open()) {
		cerr << "error: cannot access file " << filename << endl;
		exit(1);
    } // if

	// get magic number
    char magicNumber[2];
	magicNumber[0] = getChar (is);
	rmWhiteSpace (is);
	magicNumber[1] = getChar (is);
	rmWhiteSpace (is);
    if (magicNumber[0]!='P' || magicNumber[1]!='6') {
		cerr << "error: unknown file!" << endl;
		exit(1);
    } // if

    // ignore comments
    while (is->peek() == '#') {
        is->ignore(INT_MAX, '\n');
    } // while

	
    // read width and height of image
    int ww = getInt (is);
	rmWhiteSpace (is);
    int hh = getInt (is);
	rmWhiteSpace (is);

    // ignore comments
    while (is->peek() == '#') {
        is->ignore(INT_MAX, '\n');
    } // while

    // read max number of different color values
    int maxColors = getInt (is);

	// read one (!) single whitespace as delimiter!
	char tmp = getChar (is);


	// set image size
	resize(ww,hh,3);

    // read data in buffer
	char* buf = new char[width * height * 3];
    is->read (buf, width * height * 3);
    if (is->gcount() != width * height * 3) {
		cerr << "error: not enough pixels in image!" << endl;
		exit(1);
	} // if
	is->close();
	delete is;


    // copy data
    char* bufP = buf;
	for (int h=0; h<height; h++) {
		for (int w=0; w<width; w++) {
			for (int b=0; b<band; b++) {
				data[b][h][w] = *bufP;
				bufP++;
			}
		}
	}

    // delete buffer
	delete[] buf;

	return;
}


// Interface
void 
Image::write (string &filename) const
{
	if (width<1 || height<1 || band!=3 || data==NULL) {
		cerr << "error: image size not valid!" << endl;
		exit(1);
	}

	// open file
    ofstream *os = NULL;
    os = new ofstream (filename.c_str(), std::ios::out | std::ios::binary);
    if (!os || !os->good()) {
		cerr << "error: cannot open file " << filename << endl;
		exit(1);
    } // if

	*os << "P6" << endl
		<< width << " " << height << endl
		<< "255" << endl;
	

    // copy data to buffer
	char* buf = new char[width * height * 3];
	char* bufP = buf;
	for (int h=0; h<height; h++) {
		for (int w=0; w<width; w++) {
			*bufP = data[0][h][w];
			bufP++;
			*bufP = data[1][h][w];
			bufP++;
			*bufP = data[2][h][w];
			bufP++;
		}
	}

    // store buffer
	os->write(buf, width*height*3);
	os->flush();
	os->close();
	delete os;

	// delete buffer
	delete [] buf;
	
	return;
}

//  read next character 
char Image::getChar(istream *is) 
{
	decodingError=false;
	char data=0;
	is->read(&data, 1);
	if (is->gcount() != 1) {
		decodingError=true;
		//cerr << "error: no pixel available in image!" << endl;
		//exit(1);
	} // if
	return data;
}


//  read integer value 
int Image::getInt(istream *is) 
{
	decodingError=false;
	rmWhiteSpace (is);
	int v=0;

	char data=is->peek();
	bool isNumber=false;
	if ((data >= 0x30) && (data <= 0x39)) isNumber = true;

	while (isNumber) {

		// read number
		is->read(&data, 1);
		if (is->gcount() != 1) {
			decodingError=true;
			return v;
			//cerr << "error: no pixel available in image!" << endl;
			//exit(1);
		} // if

		int tmp = data - 0x30;
		v = 10*v + tmp;

		// peek next character
		data=is->peek();
		isNumber=false;
		if ((data >= 0x30) && (data <= 0x39)) isNumber = true;
	} // while
	return v;
}



// skip whitespace
void Image::rmWhiteSpace(istream *is) 
{
	decodingError=false;
	const char blank= 0x20;
	const char tab  = 0x09;
	const char CR   = 0x0d;
	const char LF   = 0x0a;

	char data=is->peek();
	bool isWs=false;
	if ((data == blank) || (data == tab) || (data == CR) || (data == LF)) isWs = true;

	while (isWs) {

		// read ws
		is->read(&data, 1);
		if (is->gcount() != 1) {
			decodingError=true;
			return;
			//cerr << "error: cannot read next byte in image" << endl;
			//exit(1);
		} // if

		// peek next character
		data=is->peek();
		isWs=false;
		if ((data == blank) || (data == tab) || (data == CR) || (data == LF)) isWs = true;
	} // while

	return;
}

void Image::allocData()
{
	freeData();

	data = new unsigned char**[band];
	for (int b=0; b<band; b++) {
		data[b] = new unsigned char*[height];
		for (int h=0; h<height; h++) {
			data[b][h] = new unsigned char[width];
		} // for
	} // for
	return;	
}


void Image::freeData()
{
	if (data != NULL) {
		for (int b=0; b<band; b++) {
			for (int h=0; h<height; h++) {
				delete[] data[b][h];
			} // for
			delete [] data[b];
		} // for
		  delete[] data;
	} // if
	data = NULL;
	return;	
}


